class ContigContainerError(Exception):
    """Exception for ContigContainer"""


class ContigError(Exception):
    """Exception for Contig"""


class RegionError(Exception):
    """Exception for Region"""


class SequenceError(Exception):
    """Exception for Sequence handling"""
